from rcmt.source import Repository


class Local(Repository):
    def __init__(self, source_name: str, project_name: str, repo_name: str):
        self.project_name = project_name
        self.repo_name = repo_name
        self.source_name = source_name

    @property
    def name(self) -> str:
        return self.repo_name

    @property
    def project(self) -> str:
        return self.project_name

    @property
    def source(self) -> str:
        return self.source_name



import datetime
import importlib.machinery
import importlib.util
import os.path
import random
import string
import sys
from typing import AnyStr, Optional

from rcmt import matcher, source
from rcmt.fs import FileProxy
from rcmt.package import action


class Run:
    """
    A Run connects packages with repositories. rcmt reads the Run, finds repositories
    and packages and then applies these packages to each repository.

    :param name: The name of the Run. rcmt uses the name to identify a run.
    :param auto_merge: rcmt automatically merges a pull request on its next run. The
                       pull request must pass all its checks.
    :param auto_merge_after: A duration after which to automatically merge a Pull
                             Request. Requires ``auto_merge`` to be set to ``true``.
    :param branch_name: Name of the branch in git. Defaults to ``branch_prefix`` +
                        ``name``.
    :param commit_msg: Message to use when committing changes via git.
    :param merge_once: If ``True``, rcmt does not create another pull request if it
                       created a pull request for the same branch before and that pull
                       request has been merged.
    :param pr_body: Define a custom body of a pull request.
    :param pr_title: Set a custom title for a pull request.

    **Example**

    .. code-block:: python

       from datetime import timedelta

       from rcmt import Run
       from rcmt.matcher import FileExists, RepoName

       with Run(
           name="python-defaults",
           auto_merge=True,
           auto_merge_after=timedelta(days=7)
       ) as run:
           run.add_matcher(FileExists("pyproject.toml"))
           run.add_matcher(RepoName("^github.com/wndhydrnt/rcmt$"))

           run.add_package("flake8")

           run.pr_title = "A custom PR title"
           run.pr_body = '''A custom PR title.
           It supports multiline strings.'''
    """

    def __init__(
        self,
        name: str,
        auto_merge: bool = False,
        auto_merge_after: Optional[datetime.timedelta] = None,
        branch_name: str = "",
        commit_msg: str = "Applied actions",
        merge_once: bool = False,
        pr_body: str = "",
        pr_title: str = "",
    ):
        self.auto_merge = auto_merge
        self.auto_merge_after = auto_merge_after
        self.branch_name = branch_name
        self.commit_msg = commit_msg
        self.pr_body = pr_body
        self.pr_title = pr_title
        self.merge_once = merge_once
        self.name = name

        self.actions: list[action.Action] = []
        self.file_proxies: list[FileProxy] = []
        self.matchers: list[matcher.Base] = []
        self.packages: list[str] = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def add_action(self, a: action.Action) -> None:
        """
        Add an Action to apply to every matching repository.

        :param a: The Action.
        """
        self.actions.append(a)

    def add_matcher(self, m: matcher.Base) -> None:
        """
        Add a Matcher that matches repositories.

        :param m: The matcher to add.
        """
        self.matchers.append(m)

    def add_package(self, name: str) -> None:
        """
        Add a Package to apply to every matching repository.

        :param name: The name of the Package.
        """
        self.packages.append(name)

    def branch(self, prefix: str) -> str:
        if self.branch_name != "":
            return self.branch_name

        return f"{prefix}{self.name}"

    def load_file(self, path: str) -> FileProxy:
        """
        Returns a proxy that an Action can use to load a file.

        :param path: Path to the file to load. Relative to the file that contains the
                     Run.
        """
        fp = FileProxy(path)
        self.file_proxies.append(fp)
        return fp

    def match(self, repo: source.Repository) -> bool:
        for m in self.matchers:
            if m.match(repo) is False:
                return False

        return True

    def set_path(self, path):
        """
        Set the path to the Run.
        Forwards this path to all ``FileProxys`` created when calling the
        ``load_file`` function.
        rcmt calls this function when it loads a Run file.

        :param path: Path to the directory that contains the Run file.
        """
        for fp in self.file_proxies:
            fp.set_path(path)


def read(path: str) -> Run:
    rndm = "".join(random.choice(string.ascii_lowercase) for _ in range(8))
    mod_name = f"rcmt_run_{rndm}"
    loader = importlib.machinery.SourceFileLoader(mod_name, path)
    spec = importlib.util.spec_from_loader(mod_name, loader)
    assert spec is not None
    new_module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = new_module
    loader.exec_module(new_module)
    try:
        run = new_module.run  # type: ignore # because the content of module is not known
        if not isinstance(run, Run):
            raise RuntimeError(
                f"Run file {path} defines variable 'run' but is not of type Run"
            )

        run.set_path(os.path.dirname(path))
        return new_module.run  # type: ignore # because the content of module is not known
    except AttributeError:
        raise RuntimeError(f"Run file {path} does not define variable 'run'")



from ..action import Action
from ..fs import FileProxy


class Manifest:
    """
    The Manifest describes the :doc:`Actions <action>` that rcmt applies to each
    repository. Every Package contains at least a Manifest file. The name of the
    Manifest file is ``manifest.py``.

    :param name: Name of the Manifest.

    **Example**

    Take a look at the `simple example <https://github.com/wndhydrnt/rcmt/tree/main/docs/examples/simple>`_
    on GitHub.
    """

    def __init__(self, name: str):
        self.name = name

        self.actions: list[Action] = []
        self.file_proxies: list[FileProxy] = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def add_action(self, action: Action) -> None:
        """
        Adds an Action. rcmt applies Actions in the order in which they are added to the
        Manifest.

        :param action: The Action to add.
        """
        self.actions.append(action)

    def load_file(self, path: str) -> FileProxy:
        """
        Returns a proxy that an Action can use to load a file.

        :param path: Path to the file to load. Relative to the file that contains the
                     Manifest.
        """
        fp = FileProxy(path)
        self.file_proxies.append(fp)
        return fp

    def set_path(self, path):
        """
        Set the path to the Package.
        Forwards this path to all ``FileProxys`` created when calling the
        ``load_file`` function.
        rcmt calls this function when it loads a Manifest file.

        :param path: Path to the directory that contains the Run file.
        """
        for fp in self.file_proxies:
            fp.set_path(path)



import io
import os
import pathlib
import re
import shutil
import string
import subprocess
import tempfile
from typing import Union

import mergedeep

from rcmt import encoding, util
from rcmt.fs import FileProxy, read_file_or_str


class Action:
    """
    Action is the abstract class that defines the interface each action implements.
    """

    def apply(self, repo_path: str, tpl_data: dict) -> None:
        """
        apply modifies a file in a repository.

        :param repo_path: The absolute path to the file in a repository to modify.
        :param tpl_data: The content of the file from a package, already populated with
                         template data.
        :return: None

        .. versionchanged:: 0.5.0
           Parameter ``pkg_path`` removed.
        """
        raise NotImplementedError("class does not implement Action.apply()")


class GlobMixin:
    """
    GlobMixin simplifies working with glob selectors.

    Classes that extend GlobMixin do not need to implement calls to ``glob`` from stdlib.
    Instead, a class implements ``process_file()`` to process each file that matches the
    glob selector.
    """

    def __init__(self, selector: str):
        self.selector = selector

    def apply(self, repo_path: str, tpl_data: dict):
        repo_file_paths = util.iglob(repo_path, self.selector)
        for repo_file_path in repo_file_paths:
            self.process_file(repo_file_path, tpl_data)

    def process_file(self, path: str, tpl_data: dict):
        raise NotImplementedError("Action does not implement GlobMixin.process_file()")


class Absent(Action):
    """
    Deletes a file or directory.

    :param target: Path to the file or directory to delete.

    **Example**

    .. code-block:: python

       # Delete the file config.yaml at the root of a repository.
       Absent(target="config.yaml")
    """

    def __init__(self, target: str):
        self.target = target

    def apply(self, repo_path: str, tpl_data: dict) -> None:
        repo_file_path = os.path.join(repo_path, self.target)
        if os.path.exists(repo_file_path):
            os.remove(repo_file_path)


class Own(Action):
    """
    Own ensures that a file in a repository stays the same.

    It always overwrites the data in the file with the data from a package.

    :param content: Content of the file to write.
    :param target: Path to the file in a repository to own.

    **Example**

    .. code-block:: python

       # Ensure that .flake8 looks the same across all repositories.
       content = "[flake8]\\nmax-line-length = 88\\nextend-ignore = E203"
       Own(content=content, target=".flake8")

    .. versionchanged:: 0.5.0
       Parameter ``content`` added. Parameter ``source`` removed.
    """

    def __init__(self, content: Union[str, FileProxy], target: str):
        self.content = content
        self.target = target

    def apply(self, repo_path: str, tpl_data: dict) -> None:
        content = read_file_or_str(self.content)
        file_path = os.path.join(repo_path, self.target)
        with open(file_path, "w+") as f:
            f.write(string.Template(content).substitute(tpl_data))


class Seed(Own):
    """
    Seed ensures that a file in a repository is present.

    It does not modify the file again if the file is present in a repository.

    :param target: Path to the file in a repository to seed.
    :param source: Path to the file in the package that contains the source data.

    **Example**

    .. code-block:: python

       # Ensure that the default Makefile is present.
       Seed(content="foo:\n\t# foo", target="Makefile")

    .. versionchanged:: 0.5.0
       Parameter ``content`` added. Parameter ``source`` removed.
    """

    def apply(self, repo_path: str, tpl_data: dict) -> None:
        repo_file_path = os.path.join(repo_path, self.target)
        if os.path.isfile(repo_file_path):
            return

        super().apply(repo_path, tpl_data)


class EncodingAware:
    def __init__(self):
        self._encodings: encoding.Registry

    @property
    def encodings(self) -> encoding.Registry:
        return self._encodings

    @encodings.setter
    def encodings(self, er: encoding.Registry):
        self._encodings = er


class Merge(Action, EncodingAware):
    """
    Merge merges the content of a file in a repository with the content of a file from a
    package.

    It supports merging of various file formats through :doc:`encoding`.

    :param selector: Glob selector to find the files to merge.
    :param source: Path to the file that contains the source data.
    :param merge_strategy: Strategy to use when merging data. ``replace`` replaces a
                           key if it already exists. ``additive`` combines collections,
                           e.g. ``list`` or ``set``.

    **Example**

    .. code-block:: python

       # Ensure that pyproject.toml contains specific keys.
       Merge(selector="pyproject.toml", source="pyproject.toml")

    .. versionchanged:: 0.5.0
       Parameter ``content`` added. Parameter ``source`` removed.
    """

    def __init__(
        self,
        content: Union[str, FileProxy],
        selector: str,
        merge_strategy: str = "replace",
    ):
        super().__init__()
        self.content = content
        self.selector = selector
        self.strategy = mergedeep.Strategy.REPLACE
        if merge_strategy == "additive":
            self.strategy = mergedeep.Strategy.ADDITIVE

    def apply(self, repo_path: str, tpl_data: dict) -> None:
        content = read_file_or_str(self.content)
        data = string.Template(content).substitute(tpl_data)
        paths = util.iglob(repo_path, self.selector)
        for p in paths:
            ext = pathlib.Path(p).suffix
            enc = self.encodings.get_for_extension(ext)
            with open(p, "r") as f:
                orig_data = enc.decode(f)

            tpl = enc.decode(io.StringIO(data))
            merged_data = enc.merge(orig_data, tpl, self.strategy)
            with open(p, "w") as f:
                enc.encode(f, merged_data)


class DeleteKey(Action, EncodingAware):
    """
    Delete a key in a file. The file has to be in a format supported by :doc:`encoding`.

    :param key: Path to the key in the data structure.
    :param target: Path to the file to modify.

    **Example**

    .. code-block:: python

       # Delete key "bar" in dict "foo" in file config.json.
       DeleteKey(key="foo.bar", target="config.json")
    """

    def __init__(self, key: str, target: str):
        super().__init__()
        self.key_path = key.split(".")
        self.target = target

    def apply(self, repo_path: str, tpl_data: dict) -> None:
        repo_file_path = os.path.join(repo_path, self.target)
        ext = pathlib.Path(repo_file_path).suffix
        enc = self.encodings.get_for_extension(ext)
        with open(repo_file_path, "r") as f:
            orig_data: dict = enc.decode(f)

        new_data = self.process_recursive(self.key_path, orig_data)
        with open(repo_file_path, "w") as f:
            enc.encode(f, new_data)

    @classmethod
    def process_recursive(cls, query: list[str], data: dict) -> dict:
        if len(query) == 0:
            return data

        qk = query[0]
        if qk not in data:
            return data

        if isinstance(data[qk], dict):
            data[qk] = cls.process_recursive(query[1:], data[qk])
        else:
            del data[qk]

        return data


class Exec(GlobMixin, Action):
    """
    Exec calls an executable and passes matching files in a repository to it. The
    executable can then modify each file.

    The executable should expect the path of a file as its only positional argument.

    :param exec_path: Path to the executable.
    :param selector: Glob selector to find the files to modify.
    :param timeout: Maximum runtime of the executable, in seconds.

    **Example**

    .. code-block:: python

       # Find all Python files in a repository recursively and pass each path to /opt/the-binary.
       Exec(exec_path="/opt/the-binary", selector="**/*.py")
    """

    def __init__(self, exec_path: str, selector: str, timeout: int = 120):
        super(Exec, self).__init__(selector)

        self.exec_path = exec_path
        self.timeout = timeout

    def process_file(self, path: str, tpl_data: dict):
        result = subprocess.run(
            args=[self.exec_path, path],
            capture_output=True,
            shell=True,
            timeout=self.timeout,
        )
        if result.returncode > 0:
            raise RuntimeError(
                f"""Exec action call to {self.exec_path} failed.
    stdout: {result.stdout.decode('utf-8')}
    stderr: {result.stderr.decode('utf-8')}"""
            )


class LineInFile(GlobMixin, Action):
    """
    LineInFile ensures that a line exists in a file. It adds the line if it does not exist.

    :param line: Line to search for.
    :param selector: Glob selector to find the files to modify.

    **Example**

    .. code-block:: python

       LineInFile(line="The Line", selector="file.txt")
    """

    def __init__(self, line: str, selector: str):
        super(LineInFile, self).__init__(selector)

        self.line = line.strip()

    def process_file(self, path: str, tpl_data: dict):
        with open(path, "r") as f:
            for line in f:
                if line.strip() == self.line:
                    return None

        with open(path, "a") as f:
            f.write(self.line)
            f.write("\n")


class DeleteLineInFile(GlobMixin, Action):
    """
    DeleteLineInFile deletes a line in a file if the file contains the line.

    :param line: Line to search for. This is a regular expression.
    :param selector: Glob selector to find the files to modify.

    **Example**

    .. code-block:: python

       DeleteLineInFile(line="The Line", selector="file.txt")

    .. versionchanged:: 0.5.0
       Does not wrap parameter ``line`` with ``^`` and ``$`` characters anymore.

    .. versionchanged:: 0.5.0
       Does not apply ``strip()`` to each line read from a file.
    """

    def __init__(self, line: str, selector: str):
        super().__init__(selector)
        self.regex = re.compile(line)

    def process_file(self, path: str, tpl_data: dict):
        with open(path, "r") as f:
            with tempfile.NamedTemporaryFile(mode="w", delete=False) as tmpf:
                tmp_file_path = tmpf.name
                line_deleted = False
                for line in f:
                    if self.regex.search(line) is None:
                        tmpf.write(line)
                    else:
                        line_deleted = True

        if line_deleted:
            shutil.move(tmp_file_path, path)


class ReplaceInLine(GlobMixin, Action):
    """
    ReplaceInLine iterates over each line of a file and replaces a string if it matches
    a regular expression.

    It uses ``re.sub()``.

    :param search: Pattern to search for. This is a regular expression.
    :param replace: Replacement if ``search`` matches.
    :param selector: Glob selector to find the files to modify.
    :param flags: Additional flags passed to ``re.sub()``.

    **Example**

    .. code-block:: python

       # Turns the line "This is a test." into "This is an example."
       ReplaceInLine(
           search="a test",
           replace=r"an example",
           selector="file.txt",
           flags=re.IGNORECASE
       )

    Both ``search`` and ``replace`` parameters support `Templating`_.

    .. note::
       Make sure to escape every ``$`` in a regular expression with ``$$`` to avoid
       errors like ``Invalid placeholder in string: line 1, col 43``.

       **Don't**

       ``^github.com/wndhydrnt/rcmt$``

       **Do**

       ``^github.com/wndhydrnt/rcmt$$``

    .. versionadded:: 0.6.0
    """

    def __init__(self, search: str, replace: str, selector: str, flags: int = 0):
        super(ReplaceInLine, self).__init__(selector)
        self.search = string.Template(search)
        self.replace = string.Template(replace)
        self.flags = flags

    def process_file(self, path: str, tpl_data: dict):
        search = self.search.substitute(tpl_data)
        replace = self.replace.substitute(tpl_data)
        with open(path, "r") as f:
            with tempfile.NamedTemporaryFile(mode="w", delete=False) as tmpf:
                for line in f:
                    tmpf.write(re.sub(search, replace, line, flags=self.flags))

        shutil.move(tmpf.name, path)



import hashlib
import importlib.machinery
import importlib.util
import os
import random
import string
import sys
from typing import Any, Hashable, Optional

import structlog

from .. import action, encoding
from .manifest import Manifest

log = structlog.get_logger()


class PackageInvalidError(RuntimeError):
    pass


class Package:
    def __init__(self, name: str):
        self.name = name
        self.actions: list[action.Action] = []


class PackageReader:
    def __init__(self, encoding_registry: encoding.Registry):
        self.encoding_registry = encoding_registry

    def read_package(self, path: str) -> Package:
        log.debug("reading package from directory", dir=path)
        manifest_path = os.path.join(path, "manifest.py")
        if not os.path.isfile(manifest_path):
            raise PackageInvalidError("manifest.py not found")

        m = load_manifest(manifest_path)
        m.set_path(os.path.basename(path))
        pkg = Package(m.name)
        for ma in m.actions:
            if isinstance(ma, action.EncodingAware):
                ma.encodings = self.encoding_registry

            pkg.actions.append(ma)

        return pkg

    def read_packages(self, paths: list[str]) -> list[Package]:
        packages = []
        for path in paths:
            log.debug("reading packages", root_dir=path)
            for entry in os.listdir(path):
                package_path = os.path.join(path, entry)
                if not os.path.isdir(package_path):
                    continue

                packages.append(self.read_package(package_path))

        return packages


def load_manifest(path: str) -> Manifest:
    rndm = "".join(random.choice(string.ascii_lowercase) for _ in range(8))
    mod_name = f"rcmt_manifest_{rndm}"
    loader = importlib.machinery.SourceFileLoader(mod_name, path)
    spec = importlib.util.spec_from_loader(mod_name, loader)
    assert spec is not None
    new_module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = new_module
    loader.exec_module(new_module)
    try:
        m = new_module.manifest  # type: ignore # because the content of module is not known
        if not isinstance(m, Manifest):
            raise RuntimeError(
                f"Run file {path} defines variable 'manifest' but is not an instance of class Manifest"
            )

        return new_module.manifest  # type: ignore # because the content of module is not known
    except AttributeError:
        raise RuntimeError(f"Manifest file {path} does not define variable 'manifest'")



import os.path
import tempfile
from typing import Any, Optional

import pydantic
import yaml
from pydantic.fields import Field


class Git(pydantic.BaseModel):
    branch_prefix: str = "rcmt/"
    clone_options: dict[str, Any] = {"filter": "blob:none"}
    data_dir: str = os.path.join(tempfile.gettempdir(), "rcmt", "data")
    user_name: str = "rcmt"
    user_email: str = ""


class Github(pydantic.BaseModel):
    access_token: str = ""
    base_url: str = "https://api.github.com"


class Gitlab(pydantic.BaseModel):
    private_token: str = ""
    url: str = "https://gitlab.com"


class Json(pydantic.BaseModel):
    indent: int = 2
    extensions: list[str] = [".json"]


class Toml(pydantic.BaseModel):
    extensions: list[str] = [".toml"]


class Yaml(pydantic.BaseModel):
    explicit_start: Optional[bool] = False
    extensions: list[str] = [".yaml", ".yml"]


class Config(pydantic.BaseSettings):
    dry_run: bool = False
    git: Git = Git()
    github: Github = Github()
    gitlab: Gitlab = Gitlab()
    # Add _ because json is a reserved field of pydantic
    json_: Json = Field(alias="json", default=Json(), env="json")
    log_level: str = "info"
    pr_title_prefix: str = "rcmt:"
    pr_title_body: str = "apply matcher {matcher_name}"
    pr_title_suffix: str = ""
    toml: Toml = Toml()
    yaml: Yaml = Yaml()

    class Config:
        env_nested_delimiter = "__"
        env_prefix = "rcmt_"
        extra = pydantic.Extra.allow


def read_config_from_file(path: str) -> Config:
    data = {}
    if path != "":
        with open(path, "r") as f:
            data = yaml.load(f, Loader=yaml.FullLoader)

    return Config(**data)



import datetime
import fnmatch
import io
import os.path
from typing import Any, TextIO, Union
from urllib.parse import urlparse

import gitlab
import structlog
from gitlab import GitlabGetError
from gitlab.base import RESTObjectList
from gitlab.v4.objects import Project as GitlabProject
from gitlab.v4.objects.merge_requests import ProjectMergeRequest as GitlabMergeRequest

from ..log import SECRET_MASKER
from .source import Base, PullRequest, Repository, add_credentials_to_url

log: structlog.stdlib.BoundLogger = structlog.get_logger(source="gitlab")


class GitlabRepository(Repository):
    def __init__(self, project: GitlabProject, token: str, url: str):
        self._project = project
        self.token = token
        self.url = url

    @property
    def base_branch(self) -> str:
        return self._project.default_branch

    @property
    def clone_url(self) -> str:
        # Value of username does not matter to GitLab.
        return add_credentials_to_url(
            url=self._project.http_url_to_repo, password=self.token, username="rcmt"
        )

    def create_pull_request(self, branch: str, pr: PullRequest) -> None:
        log.debug(
            "Creating merge request", base=self.base_branch, head=branch, repo=str(self)
        )
        self._project.mergerequests.create(
            {
                "description": pr.body,
                "source_branch": branch,
                "target_branch": self.base_branch,
                "title": pr.title,
            }
        )

    def find_pull_request(self, branch: str) -> Union[Any, None]:
        log.debug("Listing merge requests", repo=str(self))
        mrs = self._project.mergerequests.list(state="all", source_branch=branch)
        if len(mrs) == 0:
            return None

        if isinstance(mrs, list):
            return mrs[0]

        if isinstance(mrs, RESTObjectList):
            return mrs.next()

        raise RuntimeError(
            f"GitLab API returned an unexpected object {mrs.__class__.__name__}"
        )

    def get_file(self, path: str) -> TextIO:
        try:
            content = self._project.files.get(
                file_path=path, ref=self.base_branch
            ).decode()
            if content is None:
                log.warning("Decoded content is None", repo=str(self), file=path)
                raise FileNotFoundError("decoded content is None")

            return io.StringIO(content.decode("utf-8"))
        except GitlabGetError:
            raise FileNotFoundError("file does not exist in repository")

    def has_file(self, path: str) -> bool:
        directory = os.path.dirname(path)
        file = os.path.basename(path)
        tree = self._project.repository_tree(path=directory)
        for entry in tree:
            if entry["type"] != "blob":
                continue

            if fnmatch.fnmatch(entry["path"], file):
                return True

        return False

    def has_successful_pr_build(self, identifier: GitlabMergeRequest) -> bool:
        failed = False
        for commit_status in self._project.commits.get(identifier.sha).statuses.list():
            if commit_status.allow_failure is True:
                continue

            if commit_status.status != "success":
                log.debug(
                    "MR check failed",
                    repo=str(self),
                    id=identifier.get_id(),
                    name=commit_status.name,
                    status=commit_status.status,
                )
                failed = True

        if failed is False:
            log.debug(
                "All MR checks successful", repo=str(self), id=identifier.get_id()
            )
            return True

        return False

    def is_pr_closed(self, mr: GitlabMergeRequest) -> bool:
        return mr.state == "closed" or mr.state == "locked"

    def is_pr_merged(self, mr: GitlabMergeRequest) -> bool:
        return mr.state == "merged"

    def is_pr_open(self, mr: GitlabMergeRequest) -> bool:
        return mr.state == "opened"

    def merge_pull_request(self, identifier: GitlabMergeRequest) -> None:
        log.debug("Merging merge request", repo=str(self), id=identifier.get_id())
        identifier.merge()

    @property
    def name(self) -> str:
        return self._project.path

    def pr_created_at(self, pr: GitlabMergeRequest) -> datetime.datetime:
        corrected = pr.created_at.replace("Z", "+00:00")
        return datetime.datetime.fromisoformat(corrected)

    @property
    def project(self) -> str:
        return self._project.namespace["full_path"]

    @property
    def source(self) -> str:
        return self.url


class Gitlab(Base):
    def __init__(self, url: str, private_token: str):
        self.client = gitlab.Gitlab(url, private_token=private_token)
        self.url = urlparse(url).netloc

        SECRET_MASKER.add_secret(private_token)

    def list_repositories(self) -> list[Repository]:
        log.debug("start fetching repositories")
        projects = self.client.projects.list(
            all=True, archived=False, min_access_level=30
        )
        repositories: list[Repository] = []
        for p in projects:
            repositories.append(GitlabRepository(project=p, token=self.client.private_token, url=self.url))  # type: ignore

        log.debug("finished fetching repositories")
        return repositories



import os.path
from typing import Any, Mapping

import git
import structlog

from rcmt import source

log = structlog.get_logger(package="git")


class Git:
    def __init__(
        self,
        branch_name: str,
        clone_opts: Mapping[str, Any],
        data_dir: str,
        user_name: str,
        user_email: str,
    ):
        self.branch_name = branch_name
        self.clone_opts = clone_opts
        self.data_dir = data_dir

        if user_email == "":
            self.author = user_name
        else:
            self.author = f"{user_name} <{user_email}>"

    def checkout_dir(self, repo: source.Repository) -> str:
        return os.path.join(self.data_dir, repo.source, repo.project, repo.name)

    def commit_changes(self, repo_dir: str, msg: str):
        git_repo = git.Repo(path=repo_dir)
        git_repo.git.add(all=True)
        # Bug in GitPython where the type hint of `author`says the parameter is a string
        # but the underlying code expects an Actor.
        git_repo.index.commit(
            msg, author=git.objects.commit.Actor._from_string(self.author)
        )

    @staticmethod
    def has_changes(repo_dir: str) -> bool:
        git_repo = git.Repo(path=repo_dir)
        return len(git_repo.index.diff(None)) > 0 or len(git_repo.untracked_files) > 0

    def needs_push(self, repo_dir: str) -> bool:
        git_repo = git.Repo(path=repo_dir)
        try:
            return len(git_repo.index.diff(f"origin/{self.branch_name}")) > 0
        except git.BadName:
            return True

    def prepare(self, repo: source.Repository) -> str:
        checkout_dir = self.checkout_dir(repo)
        if os.path.exists(checkout_dir) is False:
            log.debug("Cloning repository", url=repo.clone_url, repo=str(repo))
            os.makedirs(checkout_dir)
            git_repo = git.Repo.clone_from(
                repo.clone_url, checkout_dir, **self.clone_opts
            )
        else:
            git_repo = git.Repo(path=checkout_dir)

        if branch_exists(self.branch_name, git_repo) is False:
            log.debug("Creating branch", branch=self.branch_name, repo=str(repo))
            git_repo.create_head(self.branch_name)

        log.debug("Checking out base branch", branch=repo.base_branch, repo=str(repo))
        git_repo.heads[repo.base_branch].checkout()
        log.debug(
            "Pulling changes into base branch", branch=repo.base_branch, repo=str(repo)
        )
        git_repo.remotes["origin"].pull()
        log.debug("Checking out work branch", branch=self.branch_name, repo=str(repo))
        git_repo.heads[self.branch_name].checkout()
        merge_base = git_repo.git.merge_base(repo.base_branch, self.branch_name)
        log.debug("Resetting to merge base", branch=self.branch_name, repo=str(repo))
        git_repo.git.reset(merge_base, hard=True)
        log.debug("Rebasing onto work branch", branch=self.branch_name, repo=str(repo))
        git_repo.git.rebase(repo.base_branch)
        return checkout_dir

    def push(self, repo_dir):
        git_repo = git.Repo(path=repo_dir)
        git_repo.git.push("origin", self.branch_name, force=True, set_upstream=True)


def branch_exists(name: str, repo: git.Repo) -> bool:
    for b in repo.heads:
        if b.name == name:
            return True

    return False



import importlib.metadata

from .matcher import FileExists as FileExistsMatcher
from .matcher import LineInFile as LineInFileMatcher
from .matcher import RepoName as RepoNameMatcher
from .rcmt import execute, execute_local, options_from_config
from .run import Run

try:
    __version__ = importlib.metadata.version(__name__)
except:
    __version__ = "develop"



import os.path
from typing import Union


class FileProxy:
    def __init__(self, file_path: str):
        self.file_path = file_path

        self.path = ""

    def set_path(self, path: str):
        self.path = path

    def read(self) -> str:
        with open(os.path.join(self.path, self.file_path), "r") as f:
            return f.read()


def read_file_or_str(content: Union[str, FileProxy]) -> str:
    if isinstance(content, FileProxy):
        return content.read()

    return content



import re

from rcmt import source


class Base:
    """
    Base class that describes the methods of a Matcher.
    """

    def match(self, repo: source.Repository) -> bool:
        """
        Indicates if a repository matches.

        :param repo: Repository to check.
        :return:
        """
        raise NotImplementedError("class does not implement Base.match()")


class FileExists(Base):
    """
    FileExists matches if the file exists in a repository.

    The code checks for the existence of a file by calling the API of the Source. This
    prevents useless checkouts of repositories and saves bandwidth.

    :param path: Path to the file, relative to the root of the repository.
                 Supports a wildcard in the name of the file, e.g. ``dir/*.json``.
    """

    def __init__(self, path: str):
        self.path = path

    def match(self, repo: source.Repository) -> bool:
        return repo.has_file(self.path)


class LineInFile(Base):
    """
    LineInFile matches if a line in a file of a repository matches a regular expression.

    It downloads the file from the repository without checking out the whole repository.

    :param path: Path of the file to check.
    :param search: Regular expression to test against each line in the file.

    .. versionadded:: 0.8.0
    """

    def __init__(self, path: str, search: str):
        self.path = path
        self.regex = re.compile(search)

    def match(self, repo: source.Repository) -> bool:
        try:
            content = repo.get_file(self.path)
            for line in content.readlines():
                if self.regex.match(line) is not None:
                    return True

            return False
        except FileNotFoundError:
            return False


class RepoName(Base):
    """
    RepoName matches if the name of a repository matches a regular expression.

    :param search: Regular expression to test against names of repositories.
    """

    def __init__(self, search: str):
        self.regex = re.compile(search)

    def match(self, repo: source.Repository) -> bool:
        if self.regex.match(str(repo)) is None:
            return False

        return True



import datetime
import logging
import sys
from typing import Optional

import structlog

from . import config, encoding, git, package, run, source
from .log import SECRET_MASKER
from .source.local import Local

structlog.configure(
    processors=[
        structlog.processors.add_log_level,
        SECRET_MASKER.process_event,
        structlog.processors.StackInfoRenderer(),
        structlog.dev.set_exc_info,
        structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M.%S", utc=False),
        structlog.dev.ConsoleRenderer(
            colors=sys.stdout is not None and sys.stdout.isatty()
        ),
    ],
    wrapper_class=structlog.make_filtering_bound_logger(logging.INFO),
)

log: structlog.stdlib.BoundLogger = structlog.get_logger()


class Options:
    def __init__(self, cfg: config.Config):
        self.config = cfg
        self.encoding_registry: encoding.Registry = encoding.Registry()
        self.matcher_path: str = ""
        self.packages_paths: list[str] = []
        self.sources: dict[str, source.Base] = {}


def can_merge_after(
    created_at: datetime.datetime, delay: Optional[datetime.timedelta]
) -> bool:
    if delay is None:
        return True

    passed = datetime.datetime.now() - created_at
    return passed >= delay


class RepoRun:
    def __init__(self, g: git.Git, opts: Options):
        self.git = g
        self.opts = opts

    def execute(
        self,
        matcher: run.Run,
        pkgs: list[package.Package],
        repo: source.Repository,
    ):
        pr_identifier = repo.find_pull_request(self.git.branch_name)
        if pr_identifier is not None and repo.is_pr_closed(pr_identifier) is True:
            log.info(
                "Existing PR has been closed",
                branch=self.git.branch_name,
                repo=str(repo),
            )
            return

        if (
            pr_identifier is not None
            and repo.is_pr_merged(pr_identifier) is True
            and matcher.merge_once
        ):
            log.info(
                "Existing PR has been merged",
                branch=self.git.branch_name,
                repo=str(repo),
            )
            return

        work_dir = self.git.prepare(repo)
        tpl_mapping = create_template_mapping(repo)
        pr = source.PullRequest(
            matcher.name,
            self.opts.config.pr_title_prefix,
            self.opts.config.pr_title_body.format(matcher_name=matcher.name),
            self.opts.config.pr_title_suffix,
            matcher.pr_body,
            matcher.pr_title,
        )
        apply_actions(pkgs, repo, matcher, tpl_mapping, work_dir)
        has_changes = False
        if self.git.has_changes(work_dir) is True:
            log.debug("Committing changes", repo=str(repo))
            self.git.commit_changes(work_dir, matcher.commit_msg)
            has_changes = True
        else:
            log.info("No changes after applying actions", repo=str(repo))

        # Combining self.git.needs_push and has_changes avoids an unnecessary push of the
        # branch if the remote branch does not exist.
        needs_push = self.git.needs_push(work_dir) and has_changes
        if needs_push:
            if self.opts.config.dry_run:
                log.warn("DRY RUN: Not pushing changes")
            else:
                log.debug("Pushing changes", repo=str(repo))
                self.git.push(work_dir)

        if needs_push is True and (
            pr_identifier is None or repo.is_pr_merged(pr_identifier) is True
        ):
            if self.opts.config.dry_run:
                log.warn("DRY RUN: Not creating pull request")
            else:
                log.info("Create pull request", repo=str(repo))
                repo.create_pull_request(self.git.branch_name, pr)

        if (
            matcher.auto_merge is True
            and needs_push is False
            and pr_identifier is not None
            and repo.is_pr_open(pr_identifier) is True
        ):
            if not repo.has_successful_pr_build(pr_identifier):
                log.warn(
                    "Cannot merge because build of pull request failed", repo=str(repo)
                )
                return

            if not can_merge_after(
                repo.pr_created_at(pr_identifier), matcher.auto_merge_after
            ):
                log.info("Too early to merge pull request", repo=str(repo))
                return

            if self.opts.config.dry_run:
                log.warn("DRY RUN: Not merging pull request", repo=str(repo))
            else:
                log.info("Merge pull request", repo=str(repo))
                repo.merge_pull_request(pr_identifier)


def apply_actions(
    pkgs: list[package.Package],
    repo: source.Repository,
    rrun: run.Run,
    tpl_mapping: dict,
    work_dir: str,
) -> None:
    for a in rrun.actions:
        log.debug(
            "Applying action from run",
            action=a.__class__.__name__,
            run=rrun.name,
            repo=str(repo),
        )
        a.apply(work_dir, tpl_mapping)

    for pkg in pkgs:
        for a in pkg.actions:
            log.debug(
                "Applying action from package",
                action=a.__class__.__name__,
                pkg=pkg.name,
                repo=str(repo),
            )
            a.apply(work_dir, tpl_mapping)


def execute(opts: Options) -> bool:
    log_level = logging.getLevelName(opts.config.log_level.upper())
    structlog.configure(
        wrapper_class=structlog.make_filtering_bound_logger(log_level),
    )

    pkg_reader = package.PackageReader(opts.encoding_registry)
    pkgs = pkg_reader.read_packages(opts.packages_paths)
    matcher = run.read(opts.matcher_path)
    pkgs_to_apply = find_packages(matcher.packages, pkgs)
    repositories = []
    for s in opts.sources.values():
        repositories += s.list_repositories()

    log.info("Repositories returned by sources", count=len(repositories))
    gitc = git.Git(
        matcher.branch(opts.config.git.branch_prefix),
        opts.config.git.clone_options,
        opts.config.git.data_dir,
        opts.config.git.user_name,
        opts.config.git.user_email,
    )
    runner = RepoRun(gitc, opts)
    success = True
    for repo in repositories:
        if matcher.match(repo) is False:
            log.debug("Repository does not match", repository=str(repo))
            continue

        log.info("Matched repository", repository=str(repo))
        try:
            runner.execute(matcher, pkgs_to_apply, repo)
        except Exception:
            log.exception("Apply failed", repository=str(repo))
            success = False

    if success is False:
        log.error("Errors during execution - check previous log messages")

    return success


def execute_local(
    directory: str, repo_source: str, repo_project: str, repo_name: str, opts: Options
) -> None:
    log_level = logging.getLevelName(opts.config.log_level.upper())
    structlog.configure(
        wrapper_class=structlog.make_filtering_bound_logger(log_level),
    )

    pkg_reader = package.PackageReader(opts.encoding_registry)
    pkgs = pkg_reader.read_packages(opts.packages_paths)
    matcher = run.read(opts.matcher_path)
    pkgs_to_apply = find_packages(matcher.packages, pkgs)
    repo = Local(repo_source, repo_project, repo_name)
    tpl_mapping: dict[str, str] = create_template_mapping(repo)
    apply_actions(pkgs_to_apply, repo, matcher, tpl_mapping, directory)


def options_from_config(path: str) -> Options:
    cfg = config.read_config_from_file(path)
    return config_to_options(cfg)


def config_to_options(cfg: config.Config) -> Options:
    opts = Options(cfg)

    opts.encoding_registry = encoding.Registry()
    opts.encoding_registry.register(
        encoding.Json(cfg.json_.indent), cfg.json_.extensions
    )
    opts.encoding_registry.register(encoding.Toml(), cfg.toml.extensions)
    opts.encoding_registry.register(
        encoding.Yaml(cfg.yaml.explicit_start), cfg.yaml.extensions
    )

    if cfg.github.access_token != "":
        source_github = source.Github(cfg.github.access_token, cfg.github.base_url)
        opts.sources["github"] = source_github

    if cfg.gitlab.private_token != "":
        source_gitlab = source.Gitlab(cfg.gitlab.url, cfg.gitlab.private_token)
        opts.sources["gitlab"] = source_gitlab

    return opts


def find_packages(
    pkg_names: list[str], pkgs: list[package.Package]
) -> list[package.Package]:
    pkgs_found = []
    pkg_names_found = []
    for pkg in pkgs:
        if pkg.name in pkg_names:
            pkg_names_found.append(pkg.name)
            pkgs_found.append(pkg)

    unknown_pkg_names = list(set(pkg_names).difference(pkg_names_found))
    if len(unknown_pkg_names) > 0:
        raise RuntimeError(f"unknown packages {unknown_pkg_names}")

    return pkgs_found


def create_template_mapping(repo: source.Repository) -> dict[str, str]:
    return {
        "repo_name": repo.name,
        "repo_project": repo.project,
        "repo_source": repo.source,
    }



"""
Encodings allow Actions to read various config file formats.
"""
import json
from typing import Any, MutableMapping, TextIO

import mergedeep
import toml
import yaml


class Encoding:
    """
    Encoding defines the API of each encoding.
    """

    def decode(self, file: TextIO) -> Any:
        """
        decode parses the content of a file-like object.

        :param file: File to parse.
        :return: Any
        """
        raise NotImplementedError("class does not implement Encoding.decode()")

    def encode(self, file: TextIO, data: dict) -> None:
        """
        encode writes data to a file-like object.

        :param file: File to write data to.
        :param data: Data to write to the file.
        :return: None
        """
        raise NotImplementedError("class does not implement Encoding.encode()")

    def merge(self, repo_data: Any, pkg_data: Any, strategy: mergedeep.Strategy) -> Any:
        """
        merge merges the data from a repository with the data from a package.

        :param repo_data: Data read from a file in a repository.
        :param pkg_data: Data read from a file in a package.
        :param strategy: The strategy to use when merging dicts. Not every encoding
                         needs to support this option.
        :return: Any
        """
        raise NotImplementedError("class does not implement Encoding.merge()")


class Registry:
    """
    Registry holds all known encodings and the file extensions each encoding supports.
    """

    def __init__(self):
        self.encodings: list[dict] = []

    def register(self, enc: Encoding, extensions: list[str]) -> None:
        """
        register registers an Encoding and its file extensions.

        :param enc: The Encoding to register.
        :param extensions: A list of file extensions, e.g. `[".yaml", ".yml"]`.
        :return: None
        """
        self.encodings.append({"encoding": enc, "extensions": extensions})

    def get_for_extension(self, ext: str) -> Encoding:
        """
        get_for_extension returns an Encoding for the given file extension.

        :param ext: The file extension to search for.
        :raises RuntimeError: If no Encoding has been registered for the file extension.
        :return: Encoding
        """
        for entry in self.encodings:
            if ext in entry["extensions"]:
                return entry["encoding"]

        raise RuntimeError(f"No encoding registered for extension {ext}")


class Json(Encoding):
    """
    Json supports json files.

    It uses Python's `json` module to decode and encode data.
    """

    def __init__(self, indent: int = 0):
        self.indent = indent

    def decode(self, file: TextIO) -> dict:
        return json.load(file)

    def encode(self, file: TextIO, data: MutableMapping) -> None:
        json_str = json.dumps(data, indent=self.indent)
        file.write(json_str + "\n")

    def merge(
        self, repo_data: dict, pkg_data: dict, strategy: mergedeep.Strategy
    ) -> MutableMapping:
        cp = repo_data.copy()
        return mergedeep.merge(cp, pkg_data, strategy=strategy)


class Yaml(Encoding):
    """
    Yaml supports yaml files.

    It uses `PyYAML` to decode and encode data.
    """

    def __init__(self, explicit_start=False):
        self.explicit_start = explicit_start

    def decode(self, file: TextIO) -> dict:
        return yaml.load(file, Loader=yaml.FullLoader)

    def encode(self, file: TextIO, data: MutableMapping) -> None:
        yaml.dump(data, file, explicit_start=self.explicit_start)

    def merge(
        self, repo_data: dict, pkg_data: dict, strategy: mergedeep.Strategy
    ) -> MutableMapping:
        cp = repo_data.copy()
        return mergedeep.merge(cp, pkg_data, strategy=strategy)


class Toml(Encoding):
    """
    Toml supports files written in Tom's Obvious, Minimal Language.
    """

    def decode(self, file: TextIO) -> MutableMapping:
        return toml.load(file)

    def encode(self, file: TextIO, data: MutableMapping) -> None:
        toml.dump(data, file)

    def merge(
        self, repo_data: dict, pkg_data: dict, strategy: mergedeep.Strategy
    ) -> MutableMapping:
        cp = repo_data.copy()
        return mergedeep.merge(cp, pkg_data, strategy=strategy)



import click

import rcmt

local_help = """Apply a Run to a local directory.

This command allows testing of a Run and its Actions quickly. It does not clone a
repository or pushes changes to a remote.

Example Usage:

rcmt local --config <config file> <Run file> <checkout of repository>
"""


@click.command(help=local_help, short_help="Apply a Run to a local directory.")
@click.option("--config", help="Path to configuration file.", default="", type=str)
@click.option(
    "--packages",
    help="Path to packages directory.",
    multiple=True,
    required=False,
    type=str,
)
@click.option("--repo-source", default="github.com", show_default=True, type=str)
@click.option("--repo-project", default="wndhydrnt", show_default=True, type=str)
@click.option("--repo-name", default="rcmt", show_default=True, type=str)
@click.argument("run_file")
@click.argument("directory")
def local(
    config: str,
    packages: list[str],
    repo_source: str,
    repo_project: str,
    repo_name: str,
    run_file: str,
    directory: str,
):
    opts = rcmt.options_from_config(config)
    opts.matcher_path = run_file
    opts.packages_paths = packages
    rcmt.execute_local(
        directory=directory,
        repo_source=repo_source,
        repo_project=repo_project,
        repo_name=repo_name,
        opts=opts,
    )


@click.command()
@click.option("--config", help="Path to configuration file.", default="", type=str)
@click.option(
    "--packages",
    help="Path to packages directory.",
    multiple=True,
    required=False,
    type=str,
)
@click.argument("matcher_file")
def run(config: str, packages: list[str], matcher_file: str):
    opts = rcmt.options_from_config(config)
    opts.matcher_path = matcher_file
    opts.packages_paths = packages
    result = rcmt.execute(opts)
    if result is False:
        exit(1)


@click.command()
def version():
    click.echo(rcmt.__version__)


@click.group()
def main():
    pass


main.add_command(local)
main.add_command(run)
main.add_command(version)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rcmt', 'rcmt.package', 'rcmt.source']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.18,<4.0.0',
 'PyGithub>=1.55,<2.0',
 'PyYAML>=5.4.1,<7.0.0',
 'click>=8.0.1,<9.0.0',
 'colorama>=0.4.4,<0.5.0',
 'mergedeep>=1.3.4,<2.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'python-gitlab>=2.10,<4.0',
 'structlog>=21.1.0,<22.0.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['rcmt = rcmt.cli:main']}

setup_kwargs = {
    'name': 'rcmt',
    'version': '0.8.1',
    'description': '',
    'long_description': '# rcmt\n\nWith rcmt you can\n\n- create, modify or delete files across many repositories.\n- merge global settings with user-configured settings in repositories.\n- write your own tooling to manipulate files in repositories.\n\nTake a look at the [documentation](https://rcmt.readthedocs.io/) to learn more.\n\n## Development\n\n```shell\n# Install dependencies and dev-dependencies\npoetry install\n# Run linters\nmake lint\n# Run tests\nmake test\n# Generate and open the documentation\nmake docs\nopen ./docs/_build/html/index.html\n```\n',
    'author': 'Markus Meyer',
    'author_email': 'hydrantanderwand@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/wndhydrnt/rcmt',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



import datetime
import urllib.parse
from typing import Any, TextIO, Union


class PullRequest:
    def __init__(
        self,
        run_name: str,
        title_prefix: str,
        title_body: str,
        title_suffix: str,
        custom_body="",
        custom_title="",
    ):
        self.custom_body = custom_body
        self.custom_title = custom_title
        self.run_name = run_name
        self.title_prefix = title_prefix
        self.title_body = title_body
        self.title_suffix = title_suffix

    def __eq__(self, other: object) -> bool:
        """
        Aids with checking equality during unit tests.

        :param other: Other PullRequest
        :return: bool
        """
        if self.custom_body != getattr(other, "custom_body"):
            return False

        if self.custom_title != getattr(other, "custom_title"):
            return False

        if self.run_name != getattr(other, "run_name"):
            return False

        if self.title_prefix != getattr(other, "title_prefix"):
            return False

        if self.title_body != getattr(other, "title_body"):
            return False

        if self.title_suffix != getattr(other, "title_suffix"):
            return False

        return True

    @property
    def body(self) -> str:
        if self.custom_body != "":
            return self.custom_body

        return f"""Apply changes from Run {self.run_name}

---

_This pull request has been created by [rcmt](https://rcmt.readthedocs.io/)._
"""

    @property
    def title(self) -> str:
        if self.custom_title != "":
            return self.custom_title

        return f"{self.title_prefix} {self.title_body} {self.title_suffix}"


class Repository:
    """
    Repository provides all methods needed to interact with a single repository of a
    Source.
    """

    def __str__(self):
        return f"{self.source}/{self.project}/{self.name}"

    @property
    def base_branch(self) -> str:
        """
        :return: Name of the base branch of this repository.
        :rtype: str
        """
        raise NotImplementedError("class does not implement Repository.base_branch()")

    @property
    def clone_url(self) -> str:
        """
        :return: Url to clone this repository.
        :rtype: str
        """
        raise NotImplementedError("class does not implement Repository.clone_url()")

    def create_pull_request(self, branch: str, pr: PullRequest) -> None:
        """
        Creates a pull request for the given branch.

        :param branch: Name of the branch.
        :param pr: The pull request.
        :rtype: None
        """
        raise NotImplementedError(
            "class does not implement Repository.create_pull_request()"
        )

    def find_pull_request(self, branch: str) -> Union[Any, None]:
        """
        Finds and returns the pull request opened by rcmt.

        :param branch: Name of the branch from which a pull request has been created.
        :return: Implementation of this method should return None if no pull request is open. Any other value will be
                 passed to merge_pull_request() so it can identify which PR to merge.
        :rtype: Any, None
        """
        raise NotImplementedError(
            "class does not implement Repository.has_open_pull_request()"
        )

    def get_file(self, path: str) -> TextIO:
        raise NotImplementedError("class does not implement Repository.has_file()")

    def has_file(self, path: str) -> bool:
        """
        Checks if a file exists in a repository.

        rcmt calls this function when matching repositories. This is more efficient than
        checking out the whole repository to check if a file exists.

        :param path: Path to a file or directory in the repository.
        :return: Indicates if the files exists.
        :rtype: bool
        """
        raise NotImplementedError("class does not implement Repository.has_file()")

    def has_successful_pr_build(self, identifier: Any) -> bool:
        """
        Checks if a pull request has passed all checks. rcmt will call merge_pull_request if this function returns
        True.

        :param identifier: Data to identify the pull request as returned by find_pull_request.
        :return: Indicates if the build is successful.
        :rtype: bool
        """
        raise NotImplementedError(
            "class does not implement Repository.has_successful_build()"
        )

    def is_pr_closed(self, identifier: Any) -> bool:
        """
        Checks if a pull request has been closed by the user.
        A pull request is closed if a user manually closes it without merging it.
        If a user has closed a pull request without merging it, rcmt will not create a
        new one.

        :param identifier: Data to identify the pull request as returned by find_pull_request.
        :return: Indicates that a pull request has been closed by the user.
        :rtype: bool
        """
        raise NotImplementedError("class does not implement Repository.is_pr_closed()")

    def is_pr_merged(self, identifier: Any) -> bool:
        """
        Checks if a pull request has been merged.
        rcmt uses this information to determine if it can create new pull requests.

        :param identifier: Data to identify the pull request as returned by find_pull_request.
        :return: Indicates that a pull request has been merged.
        :rtype: bool
        """
        raise NotImplementedError("class does not implement Repository.is_pr_merged()")

    def is_pr_open(self, identifier: Any) -> bool:
        """
        Checks if a pull request is open.
        rcmt will attempt to merge the pull request if this method returns true and
        there are no new changes from packages.

        :param identifier: Data to identify the pull request as returned by find_pull_request.
        :return: Indicates that a pull request is open.
        :rtype: bool
        """
        raise NotImplementedError("class does not implement Repository.is_pr_open()")

    def merge_pull_request(self, identifier: Any) -> None:
        """
        Merges a pull request.

        :param identifier: Data to identify the pull request as returned by find_pull_request.
        :rtype: None
        """
        raise NotImplementedError(
            "class does not implement Repository.merge_pull_request()"
        )

    @property
    def name(self) -> str:
        """
        :return: Name of the repository.
        :rtype: str
        """
        raise NotImplementedError("class does not implement Repository.name()")

    def pr_created_at(self, pr: Any) -> datetime.datetime:
        """
        Returns the date and time at which the pull request was created.

        :param pr: The pull request identifier as returned by find_pull_request().
        :return: Date and time at which the pull request was created.
        :rtype: datetime.datetime
        """
        raise NotImplementedError("class does not implement Repository.pr_created_at()")

    @property
    def project(self) -> str:
        """
        :return: Name of the project the repository belongs to.
        :rtype: str
        """

        raise NotImplementedError("class does not implement Repository.project()")

    @property
    def source(self) -> str:
        """
        :return: Name of the source that hosts the repository.
        :rtype: str
        """
        raise NotImplementedError("class does not implement Repository.source()")


class Base:
    """
    Base defines the methods every Source needs to implement.

    """

    def list_repositories(self) -> list[Repository]:
        """
        :return: List of all known repositories.
        :rtype: list[rcmt.source.Repository]
        """
        raise NotImplementedError(
            "class does not implement SourceLister.list_repositories()"
        )


def add_credentials_to_url(url: str, password: str, username: str = "") -> str:
    parsed = urllib.parse.urlparse(url)
    if username == "":
        credentials = password
    else:
        credentials = f"{username}:{password}"

    return parsed._replace(netloc=f"{credentials}@{parsed.netloc}").geturl()



import datetime
import fnmatch
import io
from typing import Any, TextIO, Union

import github
import github.PullRequest
import github.Repository
import structlog

from ..log import SECRET_MASKER
from .source import Base, PullRequest, Repository, add_credentials_to_url

log = structlog.get_logger(source="github")


class GithubRepository(Repository):
    def __init__(self, access_token: str, repo: github.Repository.Repository):
        self.access_token = access_token
        self.repo = repo
        self._name = repo.name
        self._project = repo.owner.login

    @property
    def base_branch(self) -> str:
        return self.repo.default_branch

    @property
    def clone_url(self):
        return add_credentials_to_url(
            url=self.repo.clone_url, password=self.access_token
        )

    def create_pull_request(self, branch: str, pr: PullRequest):
        log.debug(
            "Creating pull request", base=self.base_branch, head=branch, repo=str(self)
        )
        self.repo.create_pull(
            title=pr.title,
            body=pr.body,
            base=self.base_branch,
            head=branch,
            maintainer_can_modify=True,
        )

    def find_pull_request(self, branch: str) -> Union[Any, None]:
        log.debug("Listing pull requests", repo=str(self))
        for pr in self.repo.get_pulls(state="all"):
            if pr.head.ref == branch:
                return pr

        return None

    def get_file(self, path: str) -> TextIO:
        try:
            file = self.repo.get_contents(path=path, ref=self.base_branch)
        except github.UnknownObjectException:
            raise FileNotFoundError("file does not exist in repository")

        if isinstance(file, list):
            if len(file) != 1:
                raise FileNotFoundError("github returned more than one file")

            file = file[0]

        return io.StringIO(file.decoded_content.decode("utf-8"))

    def has_file(self, path: str) -> bool:
        tree = self.repo.get_git_tree(self.base_branch, True)
        for entry in tree.tree:
            if fnmatch.fnmatch(entry.path, path):
                return True

        return False

    def has_successful_pr_build(self, pr: github.PullRequest.PullRequest) -> bool:
        log.debug("Checking PR builds", repo=str(self))
        for cr in self.repo.get_commit(pr.head.sha).get_check_runs():
            if cr.conclusion != "success":
                log.debug(
                    "GitHub check not successful",
                    repo=str(self),
                    check_name=cr.name,
                    check_conclusion=cr.conclusion,
                )
                return False

        return True

    def is_pr_closed(self, pr: github.PullRequest.PullRequest) -> bool:
        return pr.state == "closed" and pr.merged is False

    def is_pr_merged(self, pr: github.PullRequest.PullRequest) -> bool:
        return pr.state == "closed" and pr.merged is True

    def is_pr_open(self, pr: github.PullRequest.PullRequest) -> bool:
        return pr.state == "open"

    def merge_pull_request(self, pr: github.PullRequest.PullRequest):
        if pr.mergeable:
            log.debug("Merging pull request", repo=str(self))
            pr.merge(commit_title="Auto-merge by rcmt")
        else:
            log.warn(
                "GitHub indicates that the PR is not mergeable",
                pr_id=pr.id,
                repo=str(self),
            )

    @property
    def name(self) -> str:
        return self.repo.name

    def pr_created_at(self, pr: github.PullRequest.PullRequest) -> datetime.datetime:
        return pr.created_at

    @property
    def project(self) -> str:
        return self.repo.owner.login

    @property
    def source(self) -> str:
        return "github.com"


class Github(Base):
    def __init__(self, access_token: str, base_url: str):
        self.access_token = access_token
        self.client = github.Github(login_or_token=access_token, base_url=base_url)

        SECRET_MASKER.add_secret(access_token)

    def list_repositories(self) -> list[Repository]:
        log.debug("start fetching repositories")
        repos: list[Repository] = []
        for gh_repo in self.client.get_user().get_repos():
            repos.append(GithubRepository(self.access_token, gh_repo))

        log.debug("finished fetching repositories")
        return repos



import glob
import os.path
from typing import Any, Generator, Union


def iglob(root: str, selector: str) -> Generator[str, Any, None]:
    """
    iglob wraps glob.iglob to check if a selector escapes the root directory. Selecting
    files outside the root directory is not allowed and results in an exception.

    It also ensures that the returned Generator contains absolute paths only.

    :param root: Root path in which to select files.
    :param selector: The selector expression to pass to glob.iglob.
    :return: Generator[str]
    """
    paths = glob.iglob(os.path.join(root, selector), recursive=True)
    for p in paths:
        abs_path = os.path.abspath(p)
        if abs_path.startswith(root) is False:
            raise RuntimeError(f"Selector {selector} escapes root directory")

        yield abs_path



from typing import Any, Mapping, MutableMapping


class SecretMasker:
    def __init__(self):
        self.secrets: list[str] = []

    def add_secret(self, s: str) -> None:
        self.secrets.append(s)

    def process_event(
        self, _, __, event_dict: MutableMapping[str, Any]
    ) -> Mapping[str, Any]:
        for k, v in event_dict.items():
            if not isinstance(v, str):
                continue

            for secret in self.secrets:
                if secret in v:
                    event_dict[k] = v.replace(secret, "****")

        return event_dict


SECRET_MASKER = SecretMasker()
